#include <stdlib.h>    // include standard libraries
#include <stdio.h>
#include <time.h>
#include <allegro.h>   // include allegro

#include "stil.h"      // include my libraries
#include "worm.h"
#include "moveplas.h"
#include "texture.h"

#define WIDTH  160
#define HEIGHT 100


DECLARE_MIDI_DRIVER_LIST()  // make .exe smaller by undeclaring allegro stuff
DECLARE_DIGI_DRIVER_LIST()
DECLARE_GFX_DRIVER_LIST (GFX_DRIVER_VGA)
DECLARE_COLOR_DEPTH_LIST (COLOR_DEPTH_8)

void   __crt0_load_environment_file(char *_app_name) { } // these three
char **__crt0_glob_function(char *_arg) { return 0; }    // commands somehow
void __crt0_setup_arguments();                           // reduce .exe size

volatile int game_time = 1;

void my_timer(void)
{
    game_time++;
}
END_OF_FUNCTION(my_timer);


// this function calls still_plasma to create the plasma, and then puts
// two copies of it onto temp2, as required by the texture-mapping and
// wormhole routines

BITMAP *create_plasma(void)
{
   BITMAP *temp, *temp2;

   temp = still_plasma(256, 256, STILL_SEAMLESS);
   temp2 = create_bitmap_ex(8, 256, 512);
   blit(temp, temp2, 0, 0, 0, 0, 256, 256);
   blit(temp, temp2, 0, 0, 0, 256, 256, 256);

   destroy_bitmap(temp);

   return temp2;
}

void main(void)
{
   PALETTE pal;
   BITMAP *stilplas, *wor1_buff, *plas_buff, *text_buff, *wor2_buff;

   plas my_plas;
   worm my_wor1, my_wor2;
   text my_text;

   float fps, y_pos = 0, x_pos = 0;
   int x, y, frame_count = 1;
   unsigned char *p;

   x = time(0); srandom(x); // initialize random seed b4 allegro installs

   allegro_init(); install_keyboard(); install_timer(); // init allegro

   LOCK_VARIABLE(game_time); LOCK_FUNCTION(my_timer);  // lock interrupt vars

   // init graphics mode

   set_gfx_mode(GFX_VGA, 320, 200, 0, 0);

   // set up a black -> blue palette

   for(x=0;x<256;x++) {
      pal[x].b = x >> 2;
      pal[x].g = 0;
      pal[x].r = 0;
   }
   set_palette(pal);

   stilplas = create_plasma();

   init_plas (&my_plas, WIDTH, HEIGHT, 4);
   init_two_worms(&my_wor1, &my_wor2, WIDTH, HEIGHT, 1000.0, 150.0);
   init_texture(&my_text, WIDTH, HEIGHT);

   wor1_buff = create_bitmap(WIDTH, HEIGHT);
   wor2_buff = create_bitmap(WIDTH, HEIGHT);
   plas_buff = create_bitmap(WIDTH, HEIGHT);
   text_buff = create_bitmap(WIDTH, HEIGHT);

   install_int(my_timer, 100);

   while(!key[KEY_ESC]) {
      fps = frame_count * 10.0 / game_time;

      y_pos += 20 / fps; x_pos += 20 / fps; // move the wormhole

      // do the effects

      plasma  (&my_plas, plas_buff, NULL, 0, fps);

      // this loop prevents the plasma looking ugly (try w/o and see)
      for(y=0;y<HEIGHT;y++) {
         p = &plas_buff->line[y][0];
         for(x=0;x<WIDTH/4;x++) {
            if(*p & 128) *p=~*p; *p = *p++ << 1;
            if(*p & 128) *p=~*p; *p = *p++ << 1;
            if(*p & 128) *p=~*p; *p = *p++ << 1;
            if(*p & 128) *p=~*p; *p = *p++ << 1;
         }
      }

      wormhole(&my_wor1, wor1_buff, stilplas, (int) y_pos, (int) x_pos);
      wormhole(&my_wor2, wor2_buff, stilplas, (int) y_pos, (int) x_pos);
      texture(&my_text, text_buff, stilplas, fps);

      // blit them to the screen

      blit(wor1_buff, screen, 1, 0, 1, 0, WIDTH-1, HEIGHT);
      blit(wor2_buff, screen, 1, 0, WIDTH+1, 0, WIDTH-1, HEIGHT);
      blit(plas_buff, screen, 0, 0, 0, HEIGHT, WIDTH, HEIGHT);
      blit(text_buff, screen, 0, 0, WIDTH, HEIGHT, WIDTH, HEIGHT);

      frame_count++;
   }
   printf("bye bye");
}
