/*
 * Particle, simple little particle simulation type program
 * By TH/5/6/97
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <dos.h>
#include <conio.h>
#include <allegro.h>
#define RANDOM(n)       (((float)rand() / (float)RAND_MAX)*(n))
#define NUMPARTICLE 3000

#pragma aux RoundToInt=\
        "fistp DWORD [eax]"\
        parm nomemory [eax] [8087]\
        modify exact [8087];

typedef struct {
        float           x;
        float           y;
        float           dirx;
        float           diry;
        int             colour;
} PARTICLE;

BITMAP                  *buffer;
unsigned char           scr_buf[64000];
int                     ytable[200];
PARTICLE                particles[NUMPARTICLE];
float                   gravity = 0.1;
PALETTE pal;


/*
 * Function: KaBoom
 * Purpose : Make an explosion at the centre-bottom of the screen
 * Input   : Nothing
 * Output  : Nothing
 */

void KaBoom(void)
{
        int             i;
        float           len, lx, ly, dist;

        for(i=0; i<NUMPARTICLE; i++) {
                particles[i].x = 160;
                particles[i].y = 180;
                particles[i].dirx = RANDOM(2) - 1;
                particles[i].diry = -RANDOM(5);
                dist = RANDOM(5);
                lx = particles[i].dirx;
                ly = particles[i].diry;
                len = sqrt(lx*lx + ly*ly);
                
                /* We need to normalize the vector to give us a direction
                 * vector, then find a random length of the vector, to give
                 * a nice, non-uniform distribution
                 */

                if(len == 0.0)
                        len = 0.0;
                else
                        len = 1.0 / len;

                particles[i].dirx *= len*dist;
                particles[i].diry *= len*dist;
                particles[i].colour = 255;
        }
}

/*
 * Function: RandomKaBoom
 * Purpose : Make a big explosion, at a random place on the screen
 * Input   : Nothing
 * Output  : Nothing
 */

void RandomKaBoom(void)
{
        int             i;
        float           len, lx, ly, dist;
        int             x, y;

        x = RANDOM(319);
        y = RANDOM(180);

        for(i=0; i<NUMPARTICLE; i++) {
                particles[i].x = x;
                particles[i].y = y;
                particles[i].dirx = RANDOM(3) - 1.5;
                particles[i].diry = RANDOM(3) - 1.5;
                dist = RANDOM(5);
                lx = particles[i].dirx;
                ly = particles[i].diry;
                len = sqrt(lx*lx + ly*ly);
                
                if(len == 0.0)
                        len = 0.0;
                else
                        len = 1.0 / len;

                particles[i].dirx *= len*dist;
                particles[i].diry *= len*dist;
                particles[i].colour = 255;
        }
}

/*
 * Function: PaintParticle
 * Purpose : Paint a single particle to the screen
 * Input   : Particle pointer
 * Output  : The particle is written to the screen
 */

void PaintParticle(PARTICLE *particle)
{
        int             x, y;

        x = (int) particle->x;
        y = (int) particle->y;

        if((x < 0) || (x > 319) || (y < 0) || (y > 199))
                return;
         scr_buf[ytable[y] + x] = particle->colour;
}

/*
 * Function: InitGraphics
 * Purpose : Set up the graphics card
 * Input   : Nothing
 * Output  : Nothing
 */

void InitGraphics(void)
{
        int             i;

        allegro_init();
        set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0);



        for(i=0; i<64; i++) {
                pal[i].r=i;
                pal[i].g=0;
                pal[i].b=0;
        }

        for(i=64; i<128; i++) {
                pal[i].r=63;
                pal[i].g=i - 64;
                pal[i].b=0;
        }

        for(i=128; i<256; i++) {
                pal[i].r=63;
                pal[i].g=63;
                pal[i].b=((i - 128) >> 1);
        }
        set_pallete(pal);

        for(i=0; i<200; i++)
                ytable[i] = i*320;

        buffer = create_bitmap(320, 200);
        clear(buffer);

}

/*
 * Function: PaintParticles
 * Purpose : Paint all of the particles in the system
 * Input   : Nothing
 * Output  : Nothing
 */

void PaintParticles(void)
{
        int             i;
        PARTICLE        *part;

        part = particles;

        for(i=0; i<NUMPARTICLE; i++) {
                PaintParticle(part);
                part++;
        }
}

/*
 * Function: MoveParticles
 * Purpose : Move all the particles in the system
 * Input   : Nothing
 * Output  : Nothing
 */

void MoveParticles(void)
{
        int             i;
        PARTICLE        *part;

        part = particles;

        for(i=0; i<NUMPARTICLE; i++) {
                part->x += part->dirx;
                part->y += part->diry;

                /* Check if the particle has collided with any of the screen
                 * top or bottom edges. If so, reflect the direction, and
                 * lose energy. If not, then just add gravity
                 */

                if(part->y > 180) {
                        part->y = 180;
                        part->dirx /= 4;
                        part->diry = -part->diry / 2;
                } else if(part->y < 1) {
                        part->y = 1;
                        part->dirx /= 4;
                        part->diry = -part->diry / 2;
                } else {
                        part->diry += gravity;
                }

                if(part->x < 0) {
                        part->x = 1;
                        part->dirx = -part->dirx / 2;
                        part->diry /= 4;
                } else if(part->x > 319) {
                        part->x = 319;
                        part->dirx = -part->dirx / 2;
                        part->diry /= 4;
                }

                /* If the particle is near the bottom of the screen, we
                 * make its colour random, to simulate the effect of the
                 * particle dying, things catching fire, that kind of stuff
                 */

                if(part->y >= 179)
                        part->colour = (rand() & 127) + 128;

                part++;
        }
}


void            blur()
{
    int             d,
                    sx,
                    sy,
                    cp,
                    cp2,
                    low_x = 1,
                    low_y = 1,
                    max_x = 319,
                    max_y = 182,
                    col_tot,
                    col;

    int row1, row2, row3;

//    memcpy(scr_buf, buffer->line[0], 64000);


    for (sy = 1; sy < 182; sy++) {
        row2=ytable[sy];
        row1=ytable[sy-1];
        row3=ytable[sy+1];
        scr_buf[row1++]=0;
        scr_buf[row2++]=0;
        scr_buf[row3++]=0;
        for (sx = 1; sx < 319; sx++) {

         col_tot=scr_buf[row1]+scr_buf[row3];
         col_tot+=scr_buf[row2-1];
         col_tot+=scr_buf[row2+1];
         col=col_tot/4;
         col-=2;
            if (col < 0)
                col = 0;
         scr_buf[row2]=col;
         row1++;
         row2++;
         row3++;
        }
        scr_buf[row1++]=0;
        scr_buf[row2++]=0;
        scr_buf[row3++]=0;
    }
   memcpy(buffer->line[0], scr_buf, 64000);

}



int main(int argc, char *argv[])
{
        int             ok = 1;

        printf("KABOOM! A great big explosion, followed by the land below");
        printf(" being burnt to a crisp\n");
        printf("Basically, it's the nearest thing to Napalm on the PC\n");
        printf("Press [K] during program to make another explosion\n");
        printf("Press any key ...\n");
        getch();

        InitGraphics();
        KaBoom();

        do {

                PaintParticles();
                MoveParticles();
                blur();
                vsync();
                blit(buffer, screen, 0, 0, 0, 0, 320, 200);
//                clear(buffer);
                if(kbhit()) {
                        switch(tolower(getch())) {
                                case 'q':
                                        ok = 0;
                                        break;
                                case 'k':
                                        RandomKaBoom();
                                        break;
                        }
                }
        } while(ok);

        return 0;
}

