#include "screenfx.h"

#include <assert.h>
#include <dos.h>
#include <time.h>
#include <values.h>

/*

   SCREENFX.C

      ScreenFX v1.1.0 18th Nov 97
      Written for DJGPP+Allegro
      Copyright (C)1997, Brett Porter

*/

static	void	pDoSpiral( int aCount1, int aCount2, int aXPos, int aYPos, BITMAP *aBitmap, int aSpiralWidth, int aSpeed, int aMasked );
static	void	pApplySine( BITMAP *aBitmap, BITMAP *aOutBitmap, fixed aAmps, fixed aComp, fixed aShift, int aMasked );

int   pInitialise()
{
   int   lCard, lWidth, lHeight;
   static RGB lWhite = { 63, 63, 63 };

	allegro_init();
	install_mouse();
	install_timer();

   set_color( 255, &lWhite );

   gui_fg_color = 0;
   gui_bg_color = 7;

	pRandomize();

   set_gfx_mode( GFX_VGA, 320, 200, 0, 0 );

   if ( !gfx_mode_select( &lCard, &lWidth, &lHeight ))
   {
      return FALSE;
   }
   return !set_gfx_mode( lCard, lWidth, lHeight, lWidth, lHeight );
}

void	pRandomize()
{
	srand( time( 0 ) << 16 );
}

int	pRandom( int aMax )
{
	long long int	lRand = rand();
	lRand *= aMax;
	return ( int )( lRand/RAND_MAX );
}

void	pSplitZapFull( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed )
{
	int		i, j, k, l, lYPos, w, h;
	BITMAP	*lOldScreen;

	w = aBitmap->w;
	h = aBitmap->h;

	lOldScreen = create_bitmap( SCREEN_W, h );
	blit(	screen, lOldScreen, 0, aYPos, 0, 0, SCREEN_W, h );

	for ( i = 0; i < h; i++ )
	{
		lYPos = i+aYPos;
		if ( i%2 == 0 )
		{
			for ( j = -aBitmap->w; j <= aXPos; j++ )
			{
				for ( k = 0; k < aBitmap->w; k++ )
				{
					while ( j+k < 0 )
					{
						k++;
					}
					if ( k < aBitmap->w )
					{
						l = aBitmap->line[i][k];
						if ( l )
						{
							putpixel( screen, j+k, lYPos, l );
						}
						else
						{
							l = lOldScreen->line[i][j+k];
							putpixel( screen, j+k, lYPos, l );
						}
					}
				}
				if ( j-1 >= 0 )
				{
					l = lOldScreen->line[i][j-1];
					putpixel( screen, j-1, i+aYPos, l );
				}
			}
		}
		else
		{
			for ( j = SCREEN_W; j >= aXPos; j-- )
			{
				for ( k = 0; k < aBitmap->w; k++ )
				{
					if ( j+k < SCREEN_W )
					{
						l = aBitmap->line[i][k];
						if ( l )
						{
							putpixel( screen, j+k, lYPos, l );
						}
						else
						{
							l = lOldScreen->line[i][j+k];
							putpixel( screen, j+k, lYPos, l );
						}
					}
				}
				if ( j+k < SCREEN_W )
				{
					l = lOldScreen->line[i][j+k];
					putpixel( screen, j+k, lYPos, l );
				}
			}
		}
		delay( aSpeed );
	}
	destroy_bitmap( lOldScreen );
}

void	pSplitZapSmall( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed )
{
	int		i, j, k, l;
	BITMAP	*lOldScreen;

	lOldScreen = create_bitmap( aBitmap->w, aBitmap->h );
	blit(	screen, lOldScreen, aXPos, aYPos, 0, 0, aBitmap->w, aBitmap->h );

	for ( i = 0; i < aBitmap->h; i++ )
	{
		if ( i%2 == 0 )
		{
			for ( j = aXPos-aBitmap->w; j <= aXPos; j++ )
			{
				for ( k = 0; k < aBitmap->w; k++ )
				{
					while ( j+k < aXPos )
					{
						k++;
					}
					if ( k < aBitmap->w )
					{
						l = aBitmap->line[i][k];
						if ( l )
						{
							putpixel( screen, j+k, i+aYPos, l );
						}
						else
						{
							l = lOldScreen->line[i][j+k-aXPos];
							putpixel( screen, j+k, i+aYPos, l );
						}
					}
				}
			}
		}
		else
		{
			for ( j = aXPos+aBitmap->w; j >= aXPos; j-- )
			{
				for ( k = 0; k < aBitmap->w; k++ )
				{
					if ( j+k < aXPos+aBitmap->w )
					{
						l = aBitmap->line[i][k];
						if ( l )
						{
							putpixel( screen, j+k, i+aYPos, l );
						}
						else
						{
							l = lOldScreen->line[i][j+k-aXPos];
							putpixel( screen, j+k, i+aYPos, l );
						}
					}
				}
			}
		}
		delay( aSpeed );
	}
	destroy_bitmap( lOldScreen );
}

void	pOverwriteZap( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
	int	i, j, k, x;

	for ( i = 0; i < aBitmap->w; i++ )
	{
		for ( j = aBitmap->h; j--; )
		{
			if ( j%2 == 0 )
			{
				x = i;
			}
			else
			{
				x = aBitmap->w-i-1;
			}
			k = aBitmap->line[j][x];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, aXPos+x, aYPos+j, k );
			}
		}
		delay( aSpeed );
	}
}

void	pDrain( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
	int	i, j, k, l;
	unsigned char	*lCurLine;
	int	lXTarget, lYTarget;
	BITMAP	*lOldScreen;

	lOldScreen = create_bitmap( aBitmap->w, aBitmap->h );
	blit(	screen, lOldScreen, aXPos, aYPos, 0, 0, aBitmap->w, aBitmap->h );

	lXTarget = aXPos+aBitmap->w;
	lYTarget = aYPos+aBitmap->h;
	for ( i = aYPos; i < lYTarget; i++ )
	{
		lCurLine = aBitmap->line[i-aYPos];
		for ( j = aXPos; j < lXTarget; j++ )
		{
			for ( k = i; k < lYTarget; k++ )
			{
				l = lCurLine[j-aXPos];
				if (( l ) || ( !aMasked ))
				{
					putpixel( screen, j, k, l );
				}
				else
				{
					l = lOldScreen->line[k-aYPos][j-aXPos];
					putpixel( screen, j, k, l );
				}
			}
		}
		delay( aSpeed );
	}
	destroy_bitmap( lOldScreen );
}

void	pBleed( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aDropletWidth, int aMasked )
{
	int	*lCounter;
	int	lLoopCheck;
	int	i, j, k, l, m;

	lCounter = malloc( aBitmap->w*sizeof( int ));
	assert( lCounter );

	if ( aDropletWidth < 1 )
 	{
		aDropletWidth = 1;
 	}
	while ( aBitmap->w%aDropletWidth )
	{
		aDropletWidth--;
	}

	for ( i = aBitmap->w/aDropletWidth; i--; )
	{
		lCounter[i] = 1;
	}

	do
	{
		lLoopCheck = TRUE;
		for ( i = ( aBitmap->w/aDropletWidth ); i--; )
		{
			j = pRandom( 10 )+1;
			if ( j+lCounter[i] > aBitmap->h )
			{
				j = aBitmap->h-lCounter[i];
			}
			for ( k = lCounter[i]-1; k < lCounter[i]+j; k++ )
			{
				for ( l = aDropletWidth; l--; )
				{
					m = aBitmap->line[k][i*aDropletWidth+l];
					if (( m ) || ( !aMasked ))
					{
						putpixel( screen, i*aDropletWidth+l+aXPos, k+aYPos, m );
					}
				}
			}
			lCounter[i] += j;
			if ( lCounter[i] < aBitmap->h )
			{
				lLoopCheck = FALSE;
			}
		}
		delay( aSpeed );
	}
	while ( !lLoopCheck );

	free( lCounter );
	lCounter = NULL;
}

void	pWipeLeftToRight( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
	int	i, j, k;

	for ( i = 0; i < aBitmap->w; i++ )
	{
		for ( j = aBitmap->h; j--; )
		{
			k = aBitmap->line[j][i];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i+aXPos, j+aYPos, k );
			}
		}
		delay( aSpeed );
	}
}

void	pWipeRightToLeft( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
	int	i, j, k;
	for ( i = aBitmap->w; i--; )
	{
		for ( j = aBitmap->h; j--; )
		{
			k = aBitmap->line[j][i];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i+aXPos, j+aYPos, k );
			}
		}
		delay( aSpeed );
	}
}

void	pWipeTopToBottom( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
	int	i, j, k;
	for ( j = 0; j < aBitmap->h; j++ )
	{
		for ( i = aBitmap->w; i--; )
		{
			k = aBitmap->line[j][i];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i+aXPos, j+aYPos, k );
			}
		}
		delay( aSpeed );
	}
}

void	pWipeBottomToTop( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
	int	i, j, k;
	for ( j = aBitmap->h; j--; )
	{
		for ( i = aBitmap->w; i--; )
		{
			k = aBitmap->line[j][i];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i+aXPos, j+aYPos, k );
			}
		}
		delay( aSpeed );
	}
}

void	pWipeToCentreVertically( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
   int   i, j, k, jtarget, oddflag;

   jtarget = aBitmap->h/2;
   oddflag = 0;

   if ( aBitmap->h%2 )
   {
      jtarget++;
      oddflag = 1;
   }

   for ( j = 0; j < jtarget; j++ )
	{
		for ( i = aBitmap->w; i--; )
		{
			k = aBitmap->line[j][i];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i+aXPos, j+aYPos, k );
			}
		}
      if (( j ) || ( !oddflag ))
      {
         for ( i = aBitmap->w; i--; )
         {
            k = aBitmap->line[aBitmap->h-j-1][i];
				if (( k ) || ( !aMasked ))
            {
               putpixel( screen, i+aXPos, aBitmap->h-j-1+aYPos, k );
            }
         }
		}
		delay( aSpeed );
	}
}

void	pWipeFromCentreVertically( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
   int   i, j, k, jtarget, oddflag;

   jtarget = aBitmap->h/2;
   oddflag = 0;

   if ( aBitmap->h%2 )
   {
      jtarget++;
      oddflag = 1;
   }

   for ( j = jtarget; j--; )
	{
		for ( i = aBitmap->w; i--; )
		{
			k = aBitmap->line[j][i];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i+aXPos, j+aYPos, k );
			}
		}
      if (( j ) || ( !oddflag ))
      {
         for ( i = aBitmap->w; i--; )
         {
            k = aBitmap->line[aBitmap->h-j-1][i];
				if (( k ) || ( !aMasked ))
            {
               putpixel( screen, i+aXPos, aBitmap->h-j-1+aYPos, k );
            }
			}
		}
		delay( aSpeed );
	}
}

void	pWipeToCentreHorizontally( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
   int   i, j, k, jtarget, oddflag;

   jtarget = aBitmap->w/2;
   oddflag = 0;

   if ( aBitmap->w%2 )
   {
      jtarget++;
      oddflag = 1;
   }

   for ( j = 0; j < jtarget; j++ )
	{
		for ( i = aBitmap->h; i--; )
		{
			k = aBitmap->line[i][j];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, j+aXPos, i+aYPos, k );
			}
		}
      if (( j ) || ( !oddflag ))
      {
         for ( i = aBitmap->h; i--; )
         {
            k = aBitmap->line[i][aBitmap->w-j-1];
				if (( k ) || ( !aMasked ))
            {
               putpixel( screen, aBitmap->w-j-1+aXPos, i+aYPos, k );
            }
         }
		}
		delay( aSpeed );
	}
}

void	pWipeFromCentreHorizontally( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed, int aMasked )
{
   int   i, j, k, jtarget, oddflag;

   jtarget = aBitmap->w/2;
   oddflag = 0;

   if ( aBitmap->w%2 )
   {
      jtarget++;
      oddflag = 1;
   }

   for ( j = jtarget; j--; )
	{
		for ( i = aBitmap->h; i--; )
		{
			k = aBitmap->line[i][j];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, j+aXPos, i+aYPos, k );
			}
		}
      if (( j ) || ( !oddflag ))
      {
         for ( i = aBitmap->h; i--; )
         {
            k = aBitmap->line[i][aBitmap->w-j-1];
				if (( k ) || ( !aMasked ))
            {
               putpixel( screen, aBitmap->w-j-1+aXPos, i+aYPos, k );
            }
         }
      }
		delay( aSpeed );
	}
}

void	pShake( int aXPos, int aYPos, BITMAP *aBitmap, int aMaxShake, int aNumShake, int aSpeed, int aMasked )
{
	int				i, lXShake, lYShake;
	BITMAP			*lOldScreen, *lBuffer;

	lBuffer = NULL;

	aXPos -= aMaxShake;
	aYPos -= aMaxShake;
	aMaxShake *= 2;

	lOldScreen = create_bitmap( aBitmap->w+aMaxShake, aBitmap->h+aMaxShake );
	blit( screen, lOldScreen, aXPos, aYPos, 0, 0, lOldScreen->w, lOldScreen->h );
	if ( aMasked )
	{
		lBuffer = create_bitmap( aBitmap->w+aMaxShake, aBitmap->h+aMaxShake );
	}

	aMaxShake--;

	lXShake = ( aMaxShake+1 )/2;	/* centre it */
	lYShake = ( aMaxShake+1 )/2;

	for ( i = aNumShake; i--; )
	{
		if ( !i )
		{
			lXShake = ( aMaxShake+1 )/2;	/* centre it again */
			lYShake = ( aMaxShake+1 )/2;
		}

		if ( aMasked )
		{
			blit( lOldScreen, lBuffer, 0, 0, 0, 0, lOldScreen->w, lOldScreen->h );
			draw_sprite( lBuffer, aBitmap, lXShake, lYShake );
			blit( lBuffer, screen, 0, 0, aXPos, aYPos, lBuffer->w, lBuffer->h );
		}
		else
		{
			blit( lOldScreen, screen, 0, 0, aXPos, aYPos, lOldScreen->w, lOldScreen->h );
			blit( aBitmap, screen, 0, 0, aXPos+lXShake, aYPos+lYShake, aBitmap->w, aBitmap->h );
		}
		delay( aSpeed );

		lXShake = pRandom( aMaxShake );
		lYShake = pRandom( aMaxShake );
	}
	destroy_bitmap( lOldScreen );
	if ( aMasked )
	{
		destroy_bitmap( lBuffer );
	}
}

void	pPushZap( int aXPos, int aYPos, BITMAP *aBitmap, int aSpeed )
{
	int	i, j;

	while ( aBitmap->w%aSpeed )
	{
		aSpeed--;
	}
	for ( i = 1; i <= aBitmap->w/aSpeed; i++ )
	{
		for ( j = 1; j <= aBitmap->h/2; j++ )
		{
			blit( screen, screen, aXPos, aYPos+j*2-1, aXPos+aSpeed, aYPos+j*2-1, aBitmap->w-aSpeed, 1 );
			blit(	aBitmap, screen, aBitmap->w-( i*aSpeed ), j*2-1, aXPos, aYPos+j*2-1, aSpeed, 1 );
		}
		for ( j = 1; j <= aBitmap->h/2; j++ )
		{
			blit( screen, screen, aXPos+aSpeed, aYPos+j*2-2, aXPos, aYPos+j*2-2, aBitmap->w-aSpeed, 1 );
			blit( aBitmap, screen, ( i-1 )*aSpeed, j*2-2, aXPos+aBitmap->w-aSpeed, aYPos+j*2-2, aSpeed, 1 );
		}
	}
}

void	pDoSpiral( int aCount1, int aCount2, int aXPos, int aYPos, BITMAP *aBitmap, int aSpiralWidth, int aSpeed, int aMasked )
{
	int	i, j, k;

	for ( i = 0; i < aBitmap->w/aSpiralWidth; i++ )
	{
		for ( j = 0; j < aBitmap->h/aSpiralWidth; j++ )
		{
			k = aBitmap->line[j*aSpiralWidth+aCount1][i*aSpiralWidth+aCount2];
			if (( k ) || ( !aMasked ))
			{
				putpixel( screen, i*aSpiralWidth+aCount2+aXPos, j*aSpiralWidth+aCount1+aYPos, k );
			}
		}
	}
	delay( aSpeed );
}

void	pSpiralOut( int aXPos, int aYPos, BITMAP *aBitmap, int aSpiralWidth, int aSpeed, int aMasked )
{
	int 	i, j;

	while (( aBitmap->w%aSpiralWidth ) || ( aBitmap->h%aSpiralWidth ))
	{
		aSpiralWidth--;
	}

	if ( aSpiralWidth == 1 )
	{
		blit( aBitmap, screen, 0, 0, aXPos, aYPos, aBitmap->w, aBitmap->h );
	}

	for ( i = ( aSpiralWidth/2 )-1; i >= 0; i-- )
	{
		for ( j = i+1; j < aSpiralWidth-i; j++ )
		{
			pDoSpiral( i, j, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
		for ( j = i; j < aSpiralWidth-i; j++ )
		{
			pDoSpiral( j, aSpiralWidth-i-1, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
		for ( j = aSpiralWidth-i-1; j >= i; j-- )
		{
			pDoSpiral( aSpiralWidth-i-1, j, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
		for ( j = aSpiralWidth-i-1; j >= i; j-- )
		{
			pDoSpiral( j, i, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
	}
}

void	pSpiralIn( int aXPos, int aYPos, BITMAP *aBitmap, int aSpiralWidth, int aSpeed, int aMasked )
{
	int 	i, j;

	while (( aBitmap->w%aSpiralWidth ) || ( aBitmap->h%aSpiralWidth ))
	{
		aSpiralWidth--;
	}

	if ( aSpiralWidth == 1 )
	{
		blit( aBitmap, screen, 0, 0, aXPos, aYPos, aBitmap->w, aBitmap->h );
	}

	for ( i = 0; i < aSpiralWidth/2; i++ )
	{
		for ( j = i; j < aSpiralWidth-i; j++ )
		{
			pDoSpiral( i, j, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
		for ( j = i; j < aSpiralWidth-i; j++ )
		{
			pDoSpiral( j, aSpiralWidth-i-1, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
		for ( j = aSpiralWidth-i-1; j >= i; j-- )
		{
			pDoSpiral( aSpiralWidth-i-1, j, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
		for ( j = aSpiralWidth-i-1; j >= i; j-- )
		{
			pDoSpiral( j, i, aXPos, aYPos, aBitmap, aSpiralWidth, aSpeed, aMasked );
		}
	}
}

void	pApplySine( BITMAP *aBitmap, BITMAP *aOutBitmap, fixed aAmps, fixed aComp, fixed aShift, int aMasked )
{
	int				lXPos, lYPos, lNewY, lDestY, lOutY;
   unsigned char	lColour, *lInLine;
	fixed				lAngle;
	const fixed		kAngleFact = 2670177L; // = 256/(PI*2)

	lDestY =	aBitmap->h-fixtoi( aAmps );

	lOutY = fixtoi( aAmps )*2;
   for( lYPos = fixtoi( aAmps ); lYPos < lDestY; lYPos++ )
   {
		lInLine = aBitmap->line[lYPos];
		lOutY++;
     	for( lXPos = aBitmap->w; lXPos--; )
      {
			lColour = lInLine[lXPos];
			if (( lColour ) || ( !aMasked ))
			{
				lAngle =	fmul( lXPos*aComp+aShift, kAngleFact );
				lNewY = fixtoi( fmul( aAmps, fsin( lAngle )))+lOutY;
				aOutBitmap->line[lNewY][lXPos] = lColour;
			}
      }
   }
}

void	pFlag( int aXPos, int aYPos, BITMAP *aBitmap, fixed aAmps, fixed aComp, int aWaves, int aSpeed, int aMasked )
{
	int		lWidth, lHeight1, lHeight2, i;
	BITMAP	*lWarp, *lOldScreen;

 	lWidth = aBitmap->w;
	lHeight1 = aBitmap->h, lHeight2 = lHeight1+2*fixtoi( aAmps );
   lWarp = create_bitmap( lWidth, lHeight2 );
	if ( aMasked )
	{
	   lOldScreen = create_bitmap( lWidth, lHeight2 );
		blit( screen, lOldScreen, aXPos, aYPos, 0, 0, lWidth, lHeight2 );
	}

   for ( i = aWaves; i--; )
   {
		if ( aMasked )
		{
			blit( lOldScreen, lWarp, 0, 0, 0, 0, lWidth, lHeight2 );
		}
		else
		{
			clear( lWarp );
		}
      pApplySine( aBitmap, lWarp, aAmps, aComp, itofix( i ), aMasked );

		blit( lWarp, screen, 0, 0, aXPos, aYPos, lWidth, lHeight2 );

      delay( aSpeed );
   }
	if ( aMasked )
	{
		destroy_bitmap( lOldScreen );
	}
	destroy_bitmap( lWarp );
}



